from bpx import BPX, Function, InterpolatedTable
import pybamm
import math
from dataclasses import dataclass
import numpy as np
from pybamm import constants
from pybamm import exp


import types
import functools


def _copy_func(f):
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


@dataclass
class Domain:
    name: str
    pre_name: str
    short_pre_name: str


cell = Domain(name="cell", pre_name="", short_pre_name="")
negative_electrode = Domain(
    name="negative electrode",
    pre_name="Negative electrode ",
    short_pre_name="Negative ",
)
positive_electrode = Domain(
    name="positive electrode",
    pre_name="Positive electrode ",
    short_pre_name="Positive ",
)
positive_current_collector = Domain(
    name="positive current collector",
    pre_name="Positive current collector ",
    short_pre_name="",
)

negative_current_collector = Domain(
    name="negative current collector",
    pre_name="Negative current collector ",
    short_pre_name="",
)

electrolyte = Domain(name="electrolyte", pre_name="Electrolyte ", short_pre_name="")
separator = Domain(name="separator", pre_name="Separator ", short_pre_name="")
experiment = Domain(name="experiment", pre_name="", short_pre_name="")


def _bpx_to_param_dict(bpx: BPX) -> dict:
    pybamm_dict = {}
    pybamm_dict = _bpx_to_domain_param_dict(
        bpx.parameterisation.cell, pybamm_dict, cell
    )
    pybamm_dict = _bpx_to_domain_param_dict(
        bpx.parameterisation.negative_electrode, pybamm_dict, negative_electrode
    )
    pybamm_dict = _bpx_to_domain_param_dict(
        bpx.parameterisation.positive_electrode, pybamm_dict, positive_electrode
    )
    pybamm_dict = _bpx_to_domain_param_dict(
        bpx.parameterisation.electrolyte, pybamm_dict, electrolyte
    )
    pybamm_dict = _bpx_to_domain_param_dict(
        bpx.parameterisation.separator, pybamm_dict, separator
    )
    pybamm_dict = _bpx_to_domain_param_dict(
        bpx.parameterisation.separator, pybamm_dict, experiment
    )

    # set a default current function and typical current based on the nominal capacity
    # i.e. a default C-rate of 1
    pybamm_dict["Current function [A]"] = pybamm_dict["Nominal cell capacity [A.h]"]

    # activity
    pybamm_dict["Thermodynamic factor"] = 1.0

    # assume Bruggeman relation for effection electrolyte properties
    for domain in [negative_electrode, separator, positive_electrode]:
        pybamm_dict[domain.pre_name + "Bruggeman coefficient (electrolyte)"] = 1.5

    # solid phase properties reported in BPX are already "effective",
    # so no correction is applied
    for domain in [negative_electrode, positive_electrode]:
        pybamm_dict[domain.pre_name + "Bruggeman coefficient (electrode)"] = 0

    # BPX is for single cell in series, user can change this later
    pybamm_dict["Number of cells connected in series to make a battery"] = 1
    pybamm_dict[
        "Number of electrodes connected in parallel to make a cell"
    ] = pybamm_dict["Number of electrode pairs connected in parallel to make a cell"]

    # electrode area
    equal_len_width = math.sqrt(pybamm_dict["Electrode area [m2]"])
    pybamm_dict["Electrode width [m]"] = equal_len_width
    pybamm_dict["Electrode height [m]"] = equal_len_width

    # surface area
    pybamm_dict["Cell cooling surface area [m2]"] = pybamm_dict[
        "External surface area [m2]"
    ]

    # volume
    pybamm_dict["Cell volume [m3]"] = pybamm_dict["Volume [m3]"]

    # reference temperature
    T_ref = pybamm_dict["Reference temperature [K]"]

    def arrhenius(Ea, T):
        return exp(Ea / constants.R * (1 / T_ref - 1 / T))

    # lumped parameters
    for name in [
        "Specific heat capacity [J.K-1.kg-1]",
        "Density [kg.m-3]",
        "Thermal conductivity [W.m-1.K-1]",
    ]:
        for domain in [
            negative_electrode,
            positive_electrode,
            separator,
            negative_current_collector,
            positive_current_collector,
        ]:
            pybamm_name = domain.pre_name + name[:1].lower() + name[1:]
            if name in pybamm_dict:
                pybamm_dict[pybamm_name] = pybamm_dict[name]

    # correct BPX specific heat capacity units to be consistent with pybamm
    for domain in [
        negative_electrode,
        positive_electrode,
        separator,
        negative_current_collector,
        positive_current_collector,
    ]:
        incorrect_name = domain.pre_name + "specific heat capacity [J.K-1.kg-1]"
        new_name = domain.pre_name + "specific heat capacity [J.kg-1.K-1]"
        if incorrect_name in pybamm_dict:
            pybamm_dict[new_name] = pybamm_dict[incorrect_name]
            del pybamm_dict[incorrect_name]

    # lumped thermal model requires current collector parameters. Arbitrarily assign
    for domain in [negative_current_collector, positive_current_collector]:
        pybamm_dict[domain.pre_name + "thickness [m]"] = 0
        pybamm_dict[domain.pre_name + "conductivity [S.m-1]"] = 4e7

    # add a default heat transfer coefficient
    pybamm_dict.update(
        {"Total heat transfer coefficient [W.m-2.K-1]": 0}, check_already_exists=False
    )

    # BET surface area
    for domain in [negative_electrode, positive_electrode]:
        pybamm_dict[domain.pre_name + "active material volume fraction"] = (
            pybamm_dict[domain.pre_name + "surface area per unit volume [m-1]"]
            * pybamm_dict[domain.short_pre_name + "particle radius [m]"]
        ) / 3.0

    # transport efficiency
    for domain in [negative_electrode, separator, positive_electrode]:
        pybamm_dict[domain.pre_name + "porosity"] = pybamm_dict[
            domain.pre_name + "transport efficiency"
        ] ** (1.0 / 1.5)

    # TODO: allow setting function parameters in a loop over domains

    # ocp

    # negative electrode (only need to check for data, other cases pass through)
    U_n = pybamm_dict[negative_electrode.pre_name + "OCP [V]"]
    if isinstance(U_n, tuple):

        def _negative_electrode_ocp(sto):
            name, (x, y) = U_n
            return pybamm.Interpolant(x, y, sto, name=name, interpolator="linear")

        pybamm_dict[negative_electrode.pre_name + "OCP [V]"] = _negative_electrode_ocp

    # positive electrode (only need to check for data, other cases pass through)
    U_p = pybamm_dict[positive_electrode.pre_name + "OCP [V]"]
    if isinstance(U_p, tuple):

        def _positive_electrode_ocp(sto):
            name, (x, y) = U_p
            return pybamm.Interpolant(x, y, sto, name=name, interpolator="linear")

        pybamm_dict[positive_electrode.pre_name + "OCP [V]"] = _positive_electrode_ocp

    # entropic change

    # negative electrode
    dUdT_n = pybamm_dict[
        negative_electrode.pre_name + "entropic change coefficient [V.K-1]"
    ]
    if callable(dUdT_n):

        def _negative_electrode_entropic_change(sto, c_s_max):
            return dUdT_n(sto)

    elif isinstance(dUdT_n, tuple):

        def _negative_electrode_entropic_change(sto, c_s_max):
            name, (x, y) = dUdT_n
            return pybamm.Interpolant(x, y, sto, name=name, interpolator="linear")

    else:

        def _negative_electrode_entropic_change(sto, c_s_max):
            return dUdT_n

    pybamm_dict[
        negative_electrode.pre_name + "OCP entropic change [V.K-1]"
    ] = _negative_electrode_entropic_change

    # positive electrode
    dUdT_p = pybamm_dict[
        positive_electrode.pre_name + "entropic change coefficient [V.K-1]"
    ]
    if callable(dUdT_p):

        def _positive_electrode_entropic_change(sto, c_s_max):
            return dUdT_p(sto)

    elif isinstance(dUdT_p, tuple):

        def _positive_electrode_entropic_change(sto, c_s_max):
            name, (x, y) = dUdT_p
            return pybamm.Interpolant(x, y, sto, name=name, interpolator="linear")

    else:

        def _positive_electrode_entropic_change(sto, c_s_max):
            return dUdT_p

    pybamm_dict[
        positive_electrode.pre_name + "OCP entropic change [V.K-1]"
    ] = _positive_electrode_entropic_change

    # reaction rates in pybamm exchange current is defined j0 = k * sqrt(ce * cs *
    # (cs-cs_max)) in BPX exchange current is defined j0 = F * k_norm * sqrt((ce/ce0) *
    # (cs/cs_max) * (1-cs/cs_max))
    c_e = pybamm_dict["Initial concentration in electrolyte [mol.m-3]"]
    F = 96485

    # negative electrode
    c_n_max = pybamm_dict[
        "Maximum concentration in " + negative_electrode.pre_name.lower() + "[mol.m-3]"
    ]
    k_n_norm = pybamm_dict[
        negative_electrode.pre_name
        + "reaction rate constant [mol.m-2.s-1]"
    ]
    Ea_k_n = pybamm_dict.get(
        negative_electrode.pre_name
        + "reaction rate constant activation energy [J.mol-1]", 0.0
    )
    # Note that in BPX j = 2*F*k_norm*sqrt((ce/ce0)*(c/c_max)*(1-c/c_max))*sinh(...),
    # and in PyBaMM j = 2*k*sqrt(ce*c*(c_max - c))*sinh(...)
    k_n = k_n_norm * F / (c_n_max * c_e**0.5)

    def _negative_electrode_exchange_current_density(c_e, c_s_surf, c_s_max, T):
        k_ref = k_n  # (A/m2)(m3/mol)**1.5 - includes ref concentrations

        return (
            k_ref
            * arrhenius(Ea_k_n, T)
            * c_e**0.5
            * c_s_surf**0.5
            * (c_s_max - c_s_surf) ** 0.5
        )

    pybamm_dict[
        negative_electrode.pre_name + "exchange-current density [A.m-2]"
    ] = _copy_func(_negative_electrode_exchange_current_density)

    # positive electrode
    c_p_max = pybamm_dict[
        "Maximum concentration in " + positive_electrode.pre_name.lower() + "[mol.m-3]"
    ]
    k_p_norm = pybamm_dict[
        positive_electrode.pre_name
        + "reaction rate constant [mol.m-2.s-1]"
    ]
    Ea_k_p = pybamm_dict.get(
        positive_electrode.pre_name
        + "reaction rate constant activation energy [J.mol-1]", 0.0
    )
    # Note that in BPX j = 2*F*k_norm*sqrt((ce/ce0)*(c/c_max)*(1-c/c_max))*sinh(...),
    # and in PyBaMM j = 2*k*sqrt(ce*c*(c_max - c))*sinh(...)
    k_p = k_p_norm * F / (c_p_max * c_e**0.5)

    def _positive_electrode_exchange_current_density(c_e, c_s_surf, c_s_max, T):
        k_ref = k_p  # (A/m2)(m3/mol)**1.5 - includes ref concentrations

        return (
            k_ref
            * arrhenius(Ea_k_p, T)
            * c_e**0.5
            * c_s_surf**0.5
            * (c_s_max - c_s_surf) ** 0.5
        )

    pybamm_dict[domain.pre_name + "exchange-current density [A.m-2]"] = _copy_func(
        _positive_electrode_exchange_current_density
    )

    # diffusivity

    # negative electrode
    Ea_D_n = pybamm_dict.get(
        negative_electrode.pre_name + "diffusivity activation energy [J.mol-1]", 0.0
    )
    D_n_ref = pybamm_dict[negative_electrode.pre_name + "diffusivity [m2.s-1]"]

    if callable(D_n_ref):

        def _negative_electrode_diffusivity(sto, T):
            return arrhenius(Ea_D_n, T) * D_n_ref(sto)

    elif isinstance(D_n_ref, tuple):

        def _negative_electrode_diffusivity(sto, T):
            name, (x, y) = D_n_ref
            return arrhenius(Ea_D_n, T) * pybamm.Interpolant(
                x, y, sto, name=name, interpolator="linear"
            )

    else:

        def _negative_electrode_diffusivity(sto, T):
            return arrhenius(Ea_D_n, T) * D_n_ref

    pybamm_dict[negative_electrode.pre_name + "diffusivity [m2.s-1]"] = _copy_func(
        _negative_electrode_diffusivity
    )

    # positive electrode
    Ea_D_p = pybamm_dict.get(
        positive_electrode.pre_name + "diffusivity activation energy [J.mol-1]", 0.0
    )
    D_p_ref = pybamm_dict[positive_electrode.pre_name + "diffusivity [m2.s-1]"]

    if callable(D_p_ref):

        def _positive_electrode_diffusivity(sto, T):
            return arrhenius(Ea_D_p, T) * D_p_ref(sto)

    elif isinstance(D_p_ref, tuple):

        def _positive_electrode_diffusivity(sto, T):
            name, (x, y) = D_p_ref
            return arrhenius(Ea_D_p, T) * pybamm.Interpolant(
                x, y, sto, name=name, interpolator="linear"
            )

    else:

        def _positive_electrode_diffusivity(sto, T):
            return arrhenius(Ea_D_p, T) * D_p_ref

    pybamm_dict[positive_electrode.pre_name + "diffusivity [m2.s-1]"] = _copy_func(
        _positive_electrode_diffusivity
    )

    # electrolyte
    Ea_D_e = pybamm_dict.get(
        electrolyte.pre_name + "diffusivity activation energy [J.mol-1]", 0.0
    )
    D_e_ref = pybamm_dict[electrolyte.pre_name + "diffusivity [m2.s-1]"]

    if callable(D_e_ref):

        def _electrolyte_diffusivity(sto, T):
            return arrhenius(Ea_D_e, T) * D_e_ref(sto)

    elif isinstance(D_e_ref, tuple):

        def _electrolyte_diffusivity(sto, T):
            name, (x, y) = D_e_ref
            return arrhenius(Ea_D_e, T) * pybamm.Interpolant(
                x, y, sto, name=name, interpolator="linear"
            )

    else:

        def _electrolyte_diffusivity(sto, T):
            return arrhenius(Ea_D_e, T) * D_e_ref

    pybamm_dict[electrolyte.pre_name + "diffusivity [m2.s-1]"] = _copy_func(
        _electrolyte_diffusivity
    )

    # conductivity
    Ea_sigma_e = pybamm_dict.get(
        electrolyte.pre_name + "conductivity activation energy [J.mol-1]", 0.0
    )
    sigma_e_ref = pybamm_dict[electrolyte.pre_name + "conductivity [S.m-1]"]

    if callable(sigma_e_ref):

        def _conductivity(c_e, T):
            return arrhenius(Ea_sigma_e, T) * sigma_e_ref(c_e)

    elif isinstance(sigma_e_ref, tuple):

        def _conductivity(c_e, T):
            name, (x, y) = sigma_e_ref
            return arrhenius(Ea_sigma_e, T) * pybamm.Interpolant(
                x, y, c_e, name=name, interpolator="linear"
            )

    else:

        def _conductivity(c_e, T):
            return arrhenius(Ea_sigma_e, T) * sigma_e_ref

    pybamm_dict[electrolyte.pre_name + "conductivity [S.m-1]"] = _copy_func(
        _conductivity
    )

    return pybamm_dict


preamble = "from pybamm import exp, tanh, cosh\n\n"


def _bpx_to_domain_param_dict(instance: BPX, pybamm_dict: dict, domain: Domain) -> dict:
    for name, field in instance.__fields__.items():
        value = getattr(instance, name)
        if value is None:
            continue
        elif isinstance(value, Function):
            value = value.to_python_function(preamble=preamble)
        elif isinstance(value, InterpolatedTable):
            # return (name, (x, y)) to match the output of
            # `pybamm.parameters.process_1D_data` we will create an interpolant on a
            # case-by-case basis to get the correct argument for each parameter
            x = np.array(value.x)
            y = np.array(value.y)
            value = (name, (x, y))

        pybamm_name = field.field_info.alias
        pybamm_name_lower = pybamm_name[:1].lower() + pybamm_name[1:]
        if pybamm_name.startswith("Initial concentration") or pybamm_name.startswith(
            "Maximum concentration"
        ):
            init_len = len("Initial concentration ")
            pybamm_name = (
                pybamm_name[:init_len]
                + "in "
                + domain.pre_name.lower()
                + pybamm_name[init_len:]
            )
        elif pybamm_name.startswith("Particle radius"):
            pybamm_name = domain.short_pre_name + pybamm_name_lower
        elif pybamm_name.startswith("OCP"):
            pybamm_name = domain.pre_name + pybamm_name
        elif pybamm_name.startswith("Cation transference number"):
            pybamm_name = pybamm_name
        elif domain.pre_name != "":
            pybamm_name = domain.pre_name + pybamm_name_lower

        pybamm_dict[pybamm_name] = value
    return pybamm_dict



#
# Base class for thermal effects
#
import pybamm


class BaseThermal(pybamm.BaseSubModel):
    """
    Base class for thermal effects

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param, options=options)

    def _get_standard_fundamental_variables(self, T_dict):
        """
        Note: here we explicitly pass in the averages for the temperature as computing
        the average temperature in `BaseThermal` using `self._x_average` requires a
        workaround to avoid raising a `ModelError` (as the key in the equation
        dict gets modified).

        For more information about this method in general,
        see :meth:`pybamm.base_submodel._get_standard_fundamental_variables`
        """
        param = self.param

        # The variable T is the concatenation of the temperature in the middle domains
        # (e.g. negative electrode, separator and positive electrode for a full cell),
        # excluding current collectors, for use in the electrochemical models
        T_mid = [T_dict[k] for k in self.options.whole_cell_domains]
        T = pybamm.concatenation(*T_mid)

        # Get the ambient temperature, which can be specified as a function of space
        # (y, z) only and time
        y = pybamm.standard_spatial_vars.y
        z = pybamm.standard_spatial_vars.z
        T_amb = param.T_amb(y, z, pybamm.t)
        T_amb_av = self._yz_average(T_amb)

        variables = {
            "Ambient temperature [K]": T_amb,
            "Volume-averaged ambient temperature [K]": T_amb_av,
            "Cell temperature [K]": T,
        }
        for name, var in T_dict.items():
            Name = name.capitalize()
            variables[f"{Name} temperature [K]"] = var
            if name in ["negative electrode", "separator", "positive electrode"]:
                variables[f"X-averaged {name} temperature [K]"] = pybamm.x_average(var)

        # Calculate temperatures in Celsius
        variables_Kelvin = variables.copy()
        for name_K, var in variables_Kelvin.items():
            name_C = name_K.replace("[K]", "[C]")
            variables.update({name_C: var - 273.15})

        return variables

    def _get_standard_coupled_variables(self, variables):
        param = self.param

        # Ohmic heating in solid
        i_s_p = variables["Positive electrode current density [A.m-2]"]
        phi_s_p = variables["Positive electrode potential [V]"]
        Q_ohm_s_cn, Q_ohm_s_cp = self._current_collector_heating(variables)
        if self.options.electrode_types["negative"] == "planar":
            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            T_n = variables["Negative electrode temperature [K]"]
            Q_ohm_s_n = i_boundary_cc**2 / param.n.sigma(T_n)
        else:
            i_s_n = variables["Negative electrode current density [A.m-2]"]
            phi_s_n = variables["Negative electrode potential [V]"]
            Q_ohm_s_n = -pybamm.inner(i_s_n, pybamm.grad(phi_s_n))
        Q_ohm_s_s = pybamm.FullBroadcast(0, ["separator"], "current collector")
        Q_ohm_s_p = -pybamm.inner(i_s_p, pybamm.grad(phi_s_p))
        Q_ohm_s = pybamm.concatenation(Q_ohm_s_n, Q_ohm_s_s, Q_ohm_s_p)

        # Ohmic heating in electrolyte
        # TODO: change full stefan-maxwell conductivity so that i_e is always
        # a Concatenation
        i_e = variables["Electrolyte current density [A.m-2]"]
        phi_e = variables["Electrolyte potential [V]"]
        if isinstance(i_e, pybamm.Concatenation):
            # compute by domain if possible
            phi_e_s = variables["Separator electrolyte potential [V]"]
            phi_e_p = variables["Positive electrolyte potential [V]"]
            if self.options.electrode_types["negative"] == "planar":
                i_e_s, i_e_p = i_e.orphans
                Q_ohm_e_n = pybamm.FullBroadcast(
                    0, ["negative electrode"], "current collector"
                )
            else:
                i_e_n, i_e_s, i_e_p = i_e.orphans
                phi_e_n = variables["Negative electrolyte potential [V]"]
                Q_ohm_e_n = -pybamm.inner(i_e_n, pybamm.grad(phi_e_n))
            Q_ohm_e_s = -pybamm.inner(i_e_s, pybamm.grad(phi_e_s))
            Q_ohm_e_p = -pybamm.inner(i_e_p, pybamm.grad(phi_e_p))
            Q_ohm_e = pybamm.concatenation(Q_ohm_e_n, Q_ohm_e_s, Q_ohm_e_p)
        else:
            # else compute using i_e across all domains
            if self.options.electrode_types["negative"] == "planar":
                Q_ohm_e_n = pybamm.FullBroadcast(
                    0, ["negative electrode"], "current collector"
                )
                Q_ohm_e_s_p = -pybamm.inner(i_e, pybamm.grad(phi_e))
                Q_ohm_e = pybamm.concatenation(Q_ohm_e_n, Q_ohm_e_s_p)
            else:
                Q_ohm_e = -pybamm.inner(i_e, pybamm.grad(phi_e))

        # Total Ohmic heating
        Q_ohm = Q_ohm_s + Q_ohm_e

        # Irreversible electrochemical heating
        a_j_p = variables[
            "Positive electrode volumetric interfacial current density [A.m-3]"
        ]
        eta_r_p = variables["Positive electrode reaction overpotential [V]"]
        if self.options.electrode_types["negative"] == "planar":
            Q_rxn_n = pybamm.FullBroadcast(
                0, ["negative electrode"], "current collector"
            )
        else:
            a_j_n = variables[
                "Negative electrode volumetric interfacial current density [A.m-3]"
            ]
            eta_r_n = variables["Negative electrode reaction overpotential [V]"]
            Q_rxn_n = a_j_n * eta_r_n
        Q_rxn_p = a_j_p * eta_r_p
        Q_rxn = pybamm.concatenation(
            Q_rxn_n, pybamm.FullBroadcast(0, "separator", "current collector"), Q_rxn_p
        )

        # Reversible electrochemical heating
        T_p = variables["Positive electrode temperature [K]"]
        dUdT_p = variables["Positive electrode entropic change [V.K-1]"]
        if self.options.electrode_types["negative"] == "planar":
            Q_rev_n = pybamm.FullBroadcast(
                0, ["negative electrode"], "current collector"
            )
        else:
            T_n = variables["Negative electrode temperature [K]"]
            dUdT_n = variables["Negative electrode entropic change [V.K-1]"]
            Q_rev_n = a_j_n * T_n * dUdT_n
        Q_rev_p = a_j_p * T_p * dUdT_p
        Q_rev = pybamm.concatenation(
            Q_rev_n, pybamm.FullBroadcast(0, "separator", "current collector"), Q_rev_p
        )

        # Total heating
        Q = Q_ohm + Q_rxn + Q_rev

        # Compute the X-average over the entire cell, including current collectors
        Q_ohm_av = self._x_average(Q_ohm, Q_ohm_s_cn, Q_ohm_s_cp)
        Q_rxn_av = self._x_average(Q_rxn, 0, 0)
        Q_rev_av = self._x_average(Q_rev, 0, 0)
        Q_av = self._x_average(Q, Q_ohm_s_cn, Q_ohm_s_cp)

        # Compute volume-averaged heat source terms
        Q_ohm_vol_av = self._yz_average(Q_ohm_av)
        Q_rxn_vol_av = self._yz_average(Q_rxn_av)
        Q_rev_vol_av = self._yz_average(Q_rev_av)
        Q_vol_av = self._yz_average(Q_av)

        variables.update(
            {
                "Ohmic heating [W.m-3]": Q_ohm,
                "X-averaged Ohmic heating [W.m-3]": Q_ohm_av,
                "Volume-averaged Ohmic heating [W.m-3]": Q_ohm_vol_av,
                "Irreversible electrochemical heating [W.m-3]": Q_rxn,
                "X-averaged irreversible electrochemical heating [W.m-3]": Q_rxn_av,
                "Volume-averaged irreversible electrochemical heating "
                + "[W.m-3]": Q_rxn_vol_av,
                "Reversible heating [W.m-3]": Q_rev,
                "X-averaged reversible heating [W.m-3]": Q_rev_av,
                "Volume-averaged reversible heating [W.m-3]": Q_rev_vol_av,
                "Total heating [W.m-3]": Q,
                "X-averaged total heating [W.m-3]": Q_av,
                "Volume-averaged total heating [W.m-3]": Q_vol_av,
                "Negative current collector Ohmic heating [W.m-3]": Q_ohm_s_cn,
                "Positive current collector Ohmic heating [W.m-3]": Q_ohm_s_cp,
            }
        )
        return variables

    def _current_collector_heating(self, variables):
        """Compute Ohmic heating in current collectors."""
        cc_dimension = self.options["dimensionality"]

        # Compute the Ohmic heating for 0D current collectors
        if cc_dimension == 0:
            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            Q_s_cn = i_boundary_cc**2 / self.param.n.sigma_cc
            Q_s_cp = i_boundary_cc**2 / self.param.p.sigma_cc
        # Otherwise we compute the Ohmic heating for 1 or 2D current collectors
        # In this limit the current flow is all in the y,z direction in the current
        # collectors
        elif cc_dimension in [1, 2]:
            phi_s_cn = variables["Negative current collector potential [V]"]
            phi_s_cp = variables["Positive current collector potential [V]"]
            # TODO: implement grad_squared in other spatial methods so that the
            # if statement can be removed
            if cc_dimension == 1:
                Q_s_cn = self.param.n.sigma_cc * pybamm.inner(
                    pybamm.grad(phi_s_cn), pybamm.grad(phi_s_cn)
                )
                Q_s_cp = self.param.p.sigma_cc * pybamm.inner(
                    pybamm.grad(phi_s_cp), pybamm.grad(phi_s_cp)
                )
            elif cc_dimension == 2:
                # Inner not implemented in 2D -- have to call grad_squared directly
                Q_s_cn = self.param.n.sigma_cc * pybamm.grad_squared(phi_s_cn)
                Q_s_cp = self.param.p.sigma_cc * pybamm.grad_squared(phi_s_cp)
        return Q_s_cn, Q_s_cp

    def _x_average(self, var, var_cn, var_cp):
        """
        Computes the X-average over the whole cell (including current collectors)
        from the variable in the cell (negative electrode, separator,
        positive electrode), negative current collector, and positive current
        collector.
        Note: we do this as we cannot create a single variable which is
        the concatenation [var_cn, var, var_cp] since var_cn and var_cp share the
        same domain. (In the N+1D formulation the current collector variables are
        assumed independent of x, so we do not make the distinction between negative
        and positive current collectors in the geometry).
        """
        out = (
            self.param.n.L_cc * var_cn
            + self.param.L_x * pybamm.x_average(var)
            + self.param.p.L_cc * var_cp
        ) / self.param.L
        return out

    def _yz_average(self, var):
        """Computes the y-z average."""
        # TODO: change the behaviour of z_average and yz_average so the if statement
        # can be removed
        if self.options["dimensionality"] in [0, 1]:
            return pybamm.z_average(var)
        elif self.options["dimensionality"] == 2:
            return pybamm.yz_average(var)
